/**
 * Copyright (c) 2010 jolira. All rights reserved. This program and the accompanying materials are made available under
 * the terms of the GNU Public License 2.0 which is available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 */

package com.google.code.joliratools.logdb.parser;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This is an immutable class for efficiently parsing log files given a set of regular expressions. This class is
 * thread-safe.
 * 
 * @author ikrimer
 * @since 1.0
 */
public class LogParser implements Parser {
    // if different instances of this class can have different patterns and formats remove static
    private static Pattern headerPattern;
    private static Pattern lineDelimiter;

    private static Collection<Pattern> patternsWithFlags = new LinkedHashSet<Pattern>();
    // TODO move following into a config file
    private static final int capacity = 1024 << 2; // should be multiple of the page size
    private static final int bufferSize = 1024 << 2; // should be multiple of the page size
    private static final String[] empty = {};

    // private static final int avgLineLength = 80;

    /**
     * Create a new parser from a set of regular expression describing the patterns to be evaluated in a log file.
     * 
     * @param regex
     *            the map of regular expression to be used for parsing to regex flags CASE_INSENSITIVE, MULTILINE,
     *            DOTALL, UNICODE_CASE, and CANON_EQ The map is linked hash map to emphasize relevance of order and
     * @param lineSeparatorRegex
     *            regular expression describing what to use as a line separator
     */
    public LogParser(final Map<String, Integer> regex, final String lineSeparatorRegex) {
        final Iterator<String> patternIterator = regex.keySet().iterator();
        String headerPatternString = patternIterator.next();
        final int headerPatternFlags = regex.get(headerPatternString).intValue();

        // lazy initialize patterns
        if (null == headerPattern || null == patternsWithFlags || null == lineDelimiter) {
            lineDelimiter = Pattern.compile(lineSeparatorRegex);
            headerPattern = Pattern.compile(headerPatternString, headerPatternFlags);
            while (patternIterator.hasNext()) {
                headerPatternString = patternIterator.next();
                patternsWithFlags.add(Pattern.compile(headerPatternString, headerPatternFlags));
            }
        }
    }

    @Override
    public Record getNext() {
        // TODO Auto-generated method stub
        return null;
    }

    private void match(final CharSequence lines) {

        for (final Pattern p : patternsWithFlags) {
            final Matcher matcher = p.matcher(lines);
            if (matcher.find()) {
                final int numberOfGroups = matcher.groupCount();
                String[] vars;
                vars = numberOfGroups > 0 ? new String[numberOfGroups + 1] : empty;
                for (int i = 0; i < numberOfGroups; i++) {
                    vars[i] = matcher.group(i);
                }
                // TODO: listener.recognized(formatter.parse(headerFields.get(0)).getTime(), headerFields.get(1),
                // headerFields.get(2), headerFields.get(3), p.pattern(), vars, lineNumberPosition);
            } else {
                // listener.unrecognized(formatter.parse(headerFields.get(0)).getTime(), headerFields.get(1),
                // headerFields.get(2), headerFields.get(3), lines.toString(), lineNumberPosition);
            }
        }
    }

    /**
     * Called to parse a log file.
     * 
     * @param in
     *            the input stream the log file should be read from
     * @param listener
     *            the listener to call when records where found
     * @throws IOException
     *             io exception
     * @throws ParseException
     *             any parsing exception
     */
    @SuppressWarnings("unused")
    private void parse(final InputStream in, final Record listener) throws IOException {
        // TODO if header is fixed replace regex with jflex
        final StringBuilder lines = new StringBuilder(capacity);
        Matcher matcher;
        Matcher newLineMatcher;
        final List<String> headerGroups = new LinkedList<String>();
        final byte[] buffer = new byte[bufferSize];
        int byteCount;
        int start, end;

        // Collection<Integer> lineBreaks = new ArrayList<Integer>(avgLineLength);
        int newLineCount = 0;

        while (0 < (byteCount = in.read(buffer))) {
            lines.append(new String(buffer, 0, byteCount - 1));
            matcher = headerPattern.matcher(lines.substring(0));

            final int[] newLineCounts = new int[lines.length()];
            newLineMatcher = lineDelimiter.matcher(lines);
            int previous = 0;
            while (newLineMatcher.find()) {
                newLineCount++;
                final int current = newLineMatcher.start() - 1;
                // this assumes that fill uses operation similar to memset
                // otherwise iterating backward to the nearest non zero would be more practical.
                Arrays.fill(newLineCounts, previous, current, newLineCount);
                previous = current;
            }

            // recordLineBreaks(lines.substring(0),lineBreaks);
            // lineNumberTotal = calculateNewLines(lines.substring(0), lineNumberTotal, false);

            // int lineNumberOfMatch=0;
            // int previousRelativeStart=0;

            int offset = 0;
            int relativeStart = 0;
            while (0 < lines.length() && matcher.find()) {
                start = matcher.start();
                end = matcher.end();

                // lineNumberOfMatch = calculateNewLines(lines.substring(1,start),lineNumberTotal, true);
                // Integer [] breakLines = new Integer[lineBreaks.size()];
                // lineBreaks.toArray(breakLines);

                final String _lines = lines.substring(0, start);

                if (0 < headerGroups.size()) {
                    match(_lines);
                    headerGroups.clear();
                } else if (0 < start) {// there is no previous header newLineCount.e. the line is junk
                    // relativeStart=start+offset;
                    // TODO: listener.junk(_lines);
                }

                // TODO if header is configurable number of groups should be configurable as well
                for (int i = 1; i <= 4; i++) {
                    headerGroups.add(matcher.group(i));
                }

                // cleanup
                relativeStart = start + offset;
                offset += end;
                lines.delete(0, end - 1);
                lines.trimToSize();
                matcher = headerPattern.matcher(lines.substring(0));
            }
        }

        // record last line

        if (0 < lines.length()) {
            // Integer [] breakLines = new Integer[lineBreaks.size()];
            // lineBreaks.toArray(breakLines);
            match(lines);// binaryFloorSearch(breakLines,0,breakLines.length,start));
        }
    }

    /*
     * private static int binaryFloorSearch(Integer [] data, int start, int end, int key){ if(start<end){ int middle =
     * (end+start)/2; if(data[middle]<key) return binaryFloorSearch(data,middle+1,end,key); else return
     * binaryFloorSearch(data,start, middle-1,key); }else{ return data[start]; } } private static void
     * recordLineBreaks(String lines, Collection<Integer> lineBreaks){ Matcher newLineMatcher; newLineMatcher =
     * lineDelimiter.matcher(lines); while(newLineMatcher.find()) lineBreaks.add(newLineMatcher.start()); } private
     * static int calculateNewLines(String lines, int lineNumberTotal, boolean matchFound) { Matcher newLineMatcher;
     * newLineMatcher = lineDelimiter.matcher(lines); int counter=matchFound?0:lineNumberTotal;
     * while(newLineMatcher.find()) counter++; return lineNumberTotal+counter; }
     */
}
